# -*- Mode: perl; indent-tabs-mode: nil -*-
#
# The contents of this file are subject to the Mozilla Public
# License Version 1.1 (the "License"); you may not use this file
# except in compliance with the License. You may obtain a copy of
# the License at http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS
# IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
# implied. See the License for the specific language governing
# rights and limitations under the License.
#
# The Original Code is the Bugzilla Bug Tracking System.
#
# The Initial Developer of the Original Code is Netscape Communications
# Corporation. Portions created by Netscape are
# Copyright (C) 1998 Netscape Communications Corporation. All
# Rights Reserved.
#
# Description: XML-RPC Interface for Bugzilla 
#
# Contributed by David Lawrence <dkl@redhat.com>
#

package Bugzilla::RPC;

## libraries and pragmas
use lib '..';
use Bug;
use Bugzilla::Constants;
use Bugzilla::Util;
use Bugzilla::Config;
use Bugzilla::User;
use Frontier::RPC2;
use Date::Parse;

use base qw(Exporter);
@Bugzilla::RPC::EXPORT = qw(handle);

use strict;
use vars qw($vars $template);

### declared variables
my $logfile = "data/xmlrpc.log";
my $debug = 0;

# Error Variables
my $RPC_FAULT_CODE;
my $RPC_FAULT_MESSAGE;

# Methods for providing bug information and submission
my $methods = {
    "bugzilla.getBug"       => \&getBug,
    "bugzilla.getLastBug"   => \&getLastBug,
    "bugzilla.getCompDesc"  => \&getCompDesc,
    "bugzilla.createBug"    => \&createBug,
    "bugzilla.getQueryInfo" => \&getQueryInfo,
    "bugzilla.runQuery"     => \&runQuery,
    "bugzilla.addComment"   => \&addComment,
    "bugzilla.updatePerms"  => \&updatePerms,
};

### Subroutines

# Shut up misguided -w warnings about "used only once". For some reason,
# "use vars" chokes on me when I try it here.
sub sillyness {
    my $zz;
    $zz = %::components;
    $zz = %::versions;
    $zz = @::legal_opsys;
    $zz = @::legal_platform;
    $zz = @::legal_priority;
    $zz = @::legal_product;
    $zz = @::legal_severity;
    # $zz = %::target_milestone;
    $zz = @::legal_resolution;
    $zz = @::legal_bug_status;
    $zz = @::target_milestone;
    $zz = @::log_columns;
    $zz = @::default_column_list;
}

sub handle {
    my ($buffer) = @_;
    Log($buffer) if $debug;
    my $rpc = new Frontier::RPC2;
    my $return = $rpc->serve($buffer, $methods);
    if ($RPC_FAULT_CODE && $RPC_FAULT_MESSAGE) {
        return $rpc->encode_fault($RPC_FAULT_CODE, $RPC_FAULT_MESSAGE);
    } else {
        return $return;
    }
}

### Section: Helper Functions

# log messages
sub Log {
    my $message = shift;
    open(LOG, ">>$logfile") || die "Could not open log file: $!";
    print LOG "$message\n";
    return;
}

sub Authenticate {
    my ($username, $password) = @_;
    my $realcryptpwd = "";
    $::userid = 0;

    # Use Bugzilla's built-in authentication to authenticate the user
    if ($username && $password) {
        # Make sure the user's login name is a valid email address.
        if (!CheckEmailSyntax($username)) {
            return 0;
        }

        # Retrieve the user's ID and crypted password from the database.
        &::SendSQL("SELECT userid, cryptpassword FROM profiles
                 WHERE login_name = " . &::SqlQuote($username));
        ($::userid, $realcryptpwd) = &::FetchSQLData();

        # Get the salt from the user's crypted password.
        my $salt = $realcryptpwd;

        # Using the salt, crypt the password the user entered.
        my $enteredCryptedPassword = crypt($password,$salt);

        # Make sure the passwords match or throw an error.
        if ($enteredCryptedPassword ne $realcryptpwd) { 
            $::userid = 0;
            return 0;
        }

        # If the user has successfully logged in, delete any password tokens
        # lying around in the system for them.
        use Token;
        my $token = Token::HasPasswordToken($::userid);
        while ( $token ) {
                Token::Cancel($token, "user logged in");
                $token = Token::HasPasswordToken($::userid);
        }
    }
   
    &::ConfirmGroup($::userid);
    $::vars->{'user'} = &::GetUserInfo($::userid);
    return $::userid;
}

sub CheckEmailSyntax {
    my ($addr) = (@_);
    my $match = Param('emailregexp');
    if ($addr !~ /$match/ || $addr =~ /[\\\(\)<>&,;:"\[\] \t\r\n]/) {
        return 0;
    }
    return 1;
}

# check and see if a given field exists, is non-empty, and is set to a
# legal value.  assume a browser bug and abort appropriately if not.
# if $legalsRef is not passed, just check to make sure the value exists and
# is non-NULL
sub CheckFormField {
    my ($formRef,                # a reference to the form to check (a hash)
        $fieldname,              # the fieldname to check
        $legalsRef               # (optional) ref to a list of legal values
       ) = @_;

    if (!defined $formRef->{$fieldname} 
        || trim($formRef->{$fieldname}) eq "" 
            || (defined($legalsRef) && lsearch($legalsRef, $formRef->{$fieldname}) < 0)) {
        return 0; 
    } else {
        return 1;
    }
}

### Section: Retrieve Bug Reports

# Return simple meta information about a bug report
sub getBug {
    my ($id, $username, $password) = @_;
    if (!$id) {
        $RPC_FAULT_CODE = 5;
        $RPC_FAULT_MESSAGE = "No bug id or alias provided.";
        return;
    }
    &::ConnectToDatabase();
    if ($username && $password) {
        Authenticate($username, $password);
    }
    my $bug = new Bug($id, $::userid);
    if ($bug->{error}) {
        $RPC_FAULT_CODE = 5;
        $RPC_FAULT_MESSAGE = $bug->{error};
        return;
    }
    return $bug->GetHash();
}

# Return last bug report submitted for product 
sub getLastBug {
    my ($product, $username, $password) = @_;
    &::ConnectToDatabase();
    if ($username && $password) {
        Authenticate($username, $password);
    }
    if (!&::CanEnterProduct($product)) {
        $RPC_FAULT_CODE = 5;
        $RPC_FAULT_MESSAGE = "Do no have permission to see product.";
        return;
    }
    if (!$product) {
        $product = "Red Hat Linux";
    }
    my $product_id = &::get_product_id($product);
    if (!$product_id) {
        $RPC_FAULT_CODE = 5;
        $RPC_FAULT_MESSAGE = "Product not found.";
        return;
    }
    &::SendSQL("SELECT bug_id FROM bugs WHERE product_id = $product_id ORDER BY bug_id DESC LIMIT 1");
    my $last_id = &::FetchOneColumn();
    my $bug = new Bug($last_id, $::userid);
    if ($bug->{error}) {
        $RPC_FAULT_CODE = 5;
        $RPC_FAULT_MESSAGE = $bug->{error};
        return;    
    }
    return $bug->GetHash();
}

### Section: Retrieve Miscellaneous Information

# Return description for component
sub getCompDesc {
    my ($product, $component, $username, $password) = @_;
    &::ConnectToDatabase();
    if ($username && $password) {
        Authenticate($username,$password);
    }
    if (!&::CanEnterProduct($product)) {
        $RPC_FAULT_CODE = 5;
        $RPC_FAULT_MESSAGE = "Do no have permission to see product.";
        return;
    }
    &::SendSQL("SELECT id FROM products WHERE name = " . &::SqlQuote($product));
    my $product_id = &::FetchOneColumn();
    if (!$product_id) {
        $RPC_FAULT_CODE = 5;
        $RPC_FAULT_MESSAGE = "Error retreiving component description. Product '$product' not found.";
        return;
    }
    &::SendSQL("SELECT description FROM components WHERE product_id = $product_id AND name = " . &::SqlQuote($component));
    if (&::MoreSQLData()) {
        return &::FetchOneColumn();
    } else {
        $RPC_FAULT_CODE = 5;
        $RPC_FAULT_MESSAGE = "Error retreiving component description. Component '$component' not found.";
        return;
    }
}

### Section: Create A Bug

sub createBug {
    my ($data, $username, $password) = @_;

    &::ConnectToDatabase();
    if ($username && $password) {
        Authenticate($username,$password);
    }
    if (!$::userid) {
        # Non-users cannot enter bugs. Account must be created first
        # through the normal interface
        $RPC_FAULT_CODE = 5;
        $RPC_FAULT_MESSAGE = "Account does not exist.";
        return;
    } 

    my $product = $data->{product};
    my $product_id = &::get_product_id($product);
    if (!$product_id) {
        $RPC_FAULT_CODE = 5;
        $RPC_FAULT_MESSAGE = "Invalid Product Name";
        return;
    }

    if (! &::CanEnterProduct($product)) {
        $RPC_FAULT_CODE = 5;
        $RPC_FAULT_MESSAGE = "Entry Access Denied for $product for User $username";
        return;
    }

    my $comment = $data->{longdesc} ? $data->{longdesc} : $data->{comment};

    if (defined($comment) && length($comment) > 65535) {
        $RPC_FAULT_CODE = 5;
        $RPC_FAULT_MESSAGE = "Comment too long (> 65535)";
        return;
    }

    my $component_id = &::get_component_id($product_id, $data->{component});
    if (!$component_id) {
        $RPC_FAULT_CODE = 5;
        $RPC_FAULT_MESSAGE = "Required Component Missing";
        return;
    }

    if (!defined $data->{'short_desc'} || trim($data->{'short_desc'}) eq "") {
        $RPC_FAULT_CODE = 5;
        $RPC_FAULT_MESSAGE = "Required Summary Missing";
        return;
    }

    $data->{'bug_file_loc'} = "" if !$data->{'bug_file_loc#'};
    
    my $sql_product = &::SqlQuote($data->{'product'});
    my $sql_component = &::SqlQuote($data->{'component'});

    # Use supplied reporter or make current userid the default
    if ($data->{'reporter'}) {
        $data->{'reporter'} = &::DBNameToIdAndCheck(trim($data->{'reporter'}));
    } else {
        $data->{'reporter'} = $::userid; 
    }
    
    # Default assignee is the component owner.
    if ($data->{'assigned_to'} eq "") {
        &::SendSQL("SELECT initialowner FROM components " .
                "WHERE id = $component_id");
        $data->{'assigned_to'} = &::FetchOneColumn();
    } else {
        $data->{'assigned_to'} = &::DBNameToIdAndCheck(trim($data->{'assigned_to'}));
    }

    my @bug_fields = ("version", "rep_platform",
                      "bug_severity", "priority", "op_sys", "assigned_to",
                      "reporter", "bug_status", "bug_file_loc", "short_desc",
                      "target_milestone");

    if (Param("useqacontact")) {
        &::SendSQL("SELECT initialqacontact FROM components " .
                "WHERE id = $component_id");
        my $qa_contact = &::FetchOneColumn();
        if (defined $qa_contact && $qa_contact != 0) {
            $data->{'qa_contact'} = $qa_contact;
            push(@bug_fields, "qa_contact");
        }
    }

    if (&::UserInGroup("canedit", $::userid) || &::UserInGroup("canconfirm", $::userid)) {
        # Default to NEW if the user hasn't selected another status
        $data->{'bug_status'} ||= "NEW";
    } else {
        # Default to UNCONFIRMED if we are using it, NEW otherwise
        $data->{'bug_status'} = $::unconfirmedstate;
        &::SendSQL("SELECT votestoconfirm FROM products WHERE id = $product_id");
        if (!&::FetchOneColumn()) {
            $data->{'bug_status'} = "NEW";
        }
    }

    if (!exists $data->{'target_milestone'}) {
        &::SendSQL("SELECT defaultmilestone FROM products WHERE name=$sql_product");
        $data->{'target_milestone'} = &::FetchOneColumn();
    }

    if (!Param('letsubmitterchoosepriority')) {
        $data->{'priority'} = Param('defaultpriority');
    }

    &::GetVersionTable();

    # Some more sanity checking
    if (!CheckFormField($data, 'product', \@::legal_product)) {
        $RPC_FAULT_CODE = 5;
        $RPC_FAULT_MESSAGE = "Illegal Value Given for product: " .  $data->{product};
        return;
    }
    if (!CheckFormField($data, 'rep_platform', \@::legal_platform)) {
        $RPC_FAULT_CODE = 5;
        $RPC_FAULT_MESSAGE = "Illegal Value Given for rep_platform: " . $data->{rep_platform}; 
        return;
    }
    if (!CheckFormField($data, 'bug_severity', \@::legal_severity)) {
        $RPC_FAULT_CODE = 5;
        $RPC_FAULT_MESSAGE = "Illegal Value Given for bug_severity: " . $data->{bug_severity};
        return;
    }
    if (!CheckFormField($data, 'priority', \@::legal_priority)) {
        $RPC_FAULT_CODE = 5;
        $RPC_FAULT_MESSAGE = "Illegal Value Given for priority: " . $data->{priority};
        return;
    }
    if (!CheckFormField($data, 'op_sys', \@::legal_opsys)) {
        $RPC_FAULT_CODE = 5;
        $RPC_FAULT_MESSAGE = "Illegal Value Given for op_sys: " . $data->{op_sys};
        return;
    }
    if (!CheckFormField($data, 'bug_status', [$::unconfirmedstate, 'NEW'])) {
        $RPC_FAULT_CODE = 5;
        $RPC_FAULT_MESSAGE = "Illegal Value Given for bug_status: " . $data->{bug_status};
        return;
    }
    if (!CheckFormField($data, 'version', $::versions{$product})) {
        $RPC_FAULT_CODE = 5;
        $RPC_FAULT_MESSAGE = "Illegal Value Given for version: " . $data->{version};
        return;
    }
    if (!CheckFormField($data, 'component', $::components{$product})) {
        $RPC_FAULT_CODE = 5;
        $RPC_FAULT_MESSAGE = "Illegal Value Given for component: " . $data->{component};
        return;
    }
    # if (!CheckFormField($data, 'target_milestone', $::target_milestone{$product})) {
    #    $RPC_FAULT_CODE = 5;
    #    $RPC_FAULT_MESSAGE = "Illegal Value Given for target_milestone: " . $data->{target_milestone};
    #    return;
    # }
    if (!defined($data->{'assigned_to'})) {
        $RPC_FAULT_CODE = 5;
        $RPC_FAULT_MESSAGE = "Illegal Value Given for assigned_to: " . $data->{assigned_to};
        return;
    }
    if (!defined($data->{'bug_file_loc'})) {
        $RPC_FAULT_CODE = 5;
        $RPC_FAULT_MESSAGE = "Illegal Value Given for bug_file_loc: " . $data->{bug_file_loc};
        return;
    }
    if (!defined($data->{'longdesc'})) {
        $RPC_FAULT_CODE = 5;
        $RPC_FAULT_MESSAGE = "Illegal Value Given for comment: " . $data->{comment};
        return;
    }

    my @used_fields;
    foreach my $field (@bug_fields) {
        if (exists $data->{$field}) {
            push (@used_fields, $field);
        }
    }

    if (exists $data->{'bug_status'} && $data->{'bug_status'} ne $::unconfirmedstate) {
        push(@used_fields, "everconfirmed");
        $data->{'everconfirmed'} = 1;
    }

    $data->{'product_id'} = $product_id;
    push(@used_fields, "product_id");
    $data->{'component_id'} = $component_id;
    push(@used_fields, "component_id");

    my %ccids;
    my @cc;

    # Create the ccid hash for inserting into the db
    # and the list for passing to processmail
    # use a hash rather than a list to avoid adding users twice
    if (defined $data->{'cc'}) {
        foreach my $person (split(/[ ,]/, $data->{'cc'})) {
            if ($person ne "") {
                my $ccid = &::DBNameToIdAndCheck($person);
                if ($ccid && !$ccids{$ccid}) {
                    $ccids{$ccid} = 1;
                    push(@cc, $person);
                }
            }
        }
    }

    # Check for valid keywords and create list of keywords to be added to db
    # (validity routine copied from process_bug.cgi)
    my @keywordlist;
    my %keywordseen;

    if ($data->{'keywords'} && &::UserInGroup("editbugs", $::userid)) {
        foreach my $keyword (split(/[\s,]+/, $data->{'keywords'})) {
            if ($keyword eq '') {
                next;
            }
            my $i = &::GetKeywordIdFromName($keyword);
            if (!$i) {
                $RPC_FAULT_CODE = 5;
                $RPC_FAULT_MESSAGE = "Unknown Keyword $keyword";
                return;
            }
            if (!$keywordseen{$i}) {
                push(@keywordlist, $i);
                $keywordseen{$i} = 1;
            }
        }
    }

    # Build up SQL string to add bug.
    my $sql = "INSERT INTO bugs " . 
    "(" . join(",", @used_fields) . ", creation_ts) " . 
    "VALUES (";

    foreach my $field (@used_fields) {
        $sql .= &::SqlQuote($data->{$field}) . ",";
    }

    $comment =~ s/\r\n?/\n/g;     # Get rid of \r.
    $comment = trim($comment);
    # If comment is all whitespace, it'll be null at this point. That's
    # OK except for the fact that it causes e-mail to be suppressed.
    $comment = $comment ? $comment : " ";

    $sql .= " now())";

    # Groups
    my @groupstoadd = ();
    if (defined $data->{'groups'} || trim($data->{'groups'}) ne "") {
        foreach my $group (split(/[\s,]+/, $data->{'groups'})) {
            &::SendSQL("SELECT id FROM groups WHERE name = " . &::SqlQuote($group));
            my $gid = &::FetchOneColumn();
            next if !$gid;
            if (!&::GroupIsActive($gid)) {
                # Prevent the user from adding the bug to an inactive group.
                # Should only happen if there is a bug in Bugzilla or the user
                # hacked the "enter bug" form since otherwise the UI 
                # for adding the bug to the group won't appear on that form.
                &::ThrowCodeError("inactive_group", undef, "abort");
            }
            &::SendSQL("SELECT user_id FROM user_group_map 
                     WHERE user_id = $::userid
                     AND group_id = $gid
                     AND isbless = 0");
            my ($permit) = &::FetchSQLData();
            if (!$permit) {
                &::SendSQL("SELECT othercontrol FROM group_control_map
                            WHERE group_id = $gid AND product_id = $product_id");
                my ($othercontrol) = &::FetchSQLData();
                $permit = (($othercontrol == CONTROLMAPSHOWN)
                           || ($othercontrol == CONTROLMAPDEFAULT));
            }
            if ($permit) {
                push(@groupstoadd, $gid)
            }
        }
    }

    &::SendSQL("SELECT DISTINCT groups.id, groups.name, " .
               "membercontrol, othercontrol " .
               "FROM groups LEFT JOIN group_control_map " .
               "ON group_id = id AND product_id = $product_id " .
               " WHERE isbuggroup != 0 AND isactive != 0 ORDER BY groups.name");
    while (&::MoreSQLData()) {
        my ($id, $groupname, $membercontrol, $othercontrol ) = &::FetchSQLData();
        $membercontrol ||= 0;
        $othercontrol ||= 0;
        # Add groups required
        if (($membercontrol == CONTROLMAPMANDATORY)
           || (($othercontrol == CONTROLMAPMANDATORY) 
                && (!&::UserInGroup($groupname)))) {
            # User had no option, bug needs to be in this group.
            push(@groupstoadd, $id)
        }
    }
    
    # Lock tables before inserting records for the new bug into the database
    # if we are using a shadow database to prevent shadow database corruption
    # when two bugs get created at the same time.
    &::SendSQL("LOCK TABLES bugs WRITE, bug_group_map WRITE, longdescs WRITE, cc WRITE, profiles READ") if Param("shadowdb");

    # Add the bug report to the DB.
    &::SendSQL($sql);

    # Get the bug ID back.
    my $id = &::CurrId("bugs:bug_id");

    # Add the group restrictions
    foreach my $grouptoadd (@groupstoadd) {
        &::SendSQL("INSERT INTO bug_group_map (bug_id, group_id)
                VALUES ($id, $grouptoadd)");
    }

    # Add the comment
    &::SendSQL("INSERT INTO longdescs (bug_id, who, bug_when, thetext) 
            VALUES ($id, $::userid, now(), " . &::SqlQuote($comment) . ")");

    # Insert the cclist into the database
    foreach my $ccid (keys(%ccids)) {
        &::SendSQL("INSERT INTO cc (bug_id, who) VALUES ($id, $ccid)");
    }

    if (&::UserInGroup("editbugs")) {
        foreach my $keyword (@keywordlist) {
            &::SendSQL("INSERT INTO keywords (bug_id, keywordid) 
                     VALUES ($id, $keyword)");
        }
    }

    &::SendSQL("UNLOCK TABLES") if Param("shadowdb");

    # Assemble the -force* strings so this counts as "Added to this capacity"
    my @ARGLIST = ();
    if (@cc) {
        push (@ARGLIST, "-forcecc", join(",", @cc));
    }

    push (@ARGLIST, "-forceowner", &::DBID_to_name($data->{assigned_to}));


    if (defined $data->{'qa_contact'}) {
        push (@ARGLIST, "-forceqacontact", &::DBID_to_name($data->{'qa_contact'}));
    }

    push (@ARGLIST, "-forcereporter", &::DBID_to_name($::userid));

    push (@ARGLIST, $id, $username);

    # Send mail to let people know the bug has been created.
    # See attachment.cgi for explanation of why it's done this way.
    my $mailresults = '';
    open(PMAIL, "-|") or exec('./processmail', @ARGLIST);
    $mailresults .= $_ while <PMAIL>;
    close(PMAIL);

    return $id;
}

### Section: Add Comment

sub addComment {
    my ($id, $comment, $username, $password, $isprivate,  $timestamp, $work_time) = @_;
    if (!$id || !$username || !$password) {
        $RPC_FAULT_CODE = 5;
        $RPC_FAULT_MESSAGE = "No bug id or alias provided or did not provide proper login information.";
        return;
    }
    if (!$comment) {
        $RPC_FAULT_CODE = 5;
        $RPC_FAULT_MESSAGE = "No comment text provided.";
        return;
    }
    $isprivate ||= 0;
    
    &::ConnectToDatabase();
    if ($username && $password) {
        Authenticate($username, $password);
    }

    &::AppendComment($id, $username, $comment, $isprivate, $timestamp, $work_time);

    # Assemble the -force* strings so this counts as "Added to this capacity"
    my @ARGLIST = ();
    push (@ARGLIST, $id, $username);
                                                                                                                                                               
    # Send mail to let people know the bug has been created.
    # See attachment.cgi for explanation of why it's done this way.
    my $mailresults = '';
    open(PMAIL, "-|") or exec('./processmail', @ARGLIST);
    $mailresults .= $_ while <PMAIL>;
    close(PMAIL);
    
    return 1;
}

### Section: Update Perms 

sub updatePerms {
    my ($user, $cmd, $groups, $username, $password) = @_;
    my $return = "";

    if (!$user || !$cmd || !$username || !$password) {
        $RPC_FAULT_CODE = 5;
        $RPC_FAULT_MESSAGE = "No username provided or did not provide proper login information.";
        return;
    }

    &::ConnectToDatabase();
    if ($username && $password) {
        Authenticate($username, $password);
    }

    if (! &::UserInGroup('editusers')) {
        return "Error: you do not have edit member permissions.";
    }

    # Return if user does not exist
    my $userid = &::DBname_to_id($user);
    if (!$userid) {
        return "Error: User $user does not exist.";
    }

    if ($cmd eq 'list') {
        my @groups;
        my $query = "
        SELECT 
            DISTINCT groups.name
        FROM 
            groups JOIN user_group_map ON groups.id = user_group_map.group_id
        WHERE
            user_group_map.user_id = $userid 
            AND isbless = 0";
        &::SendSQL($query);
        while (my @row = &::FetchSQLData()) {
            push (@groups, $row[0]);
        }
        return join(", ", @groups) if @groups;  
        return "No groups found";
    }   
       
    my @added;
    my @removed;
    foreach my $group (split(/ /, $groups)) {
        my $query = "SELECT id FROM groups WHERE name = " . &::SqlQuote($group);
        &::SendSQL($query); 
        my $groupid = &::FetchOneColumn();
        next if !$groupid;

        if ($cmd =~ 'add') {
            my $query = "SELECT group_id FROM user_group_map WHERE user_id = $userid " .
                        "AND group_id = $groupid AND isbless = 0";
            &::SendSQL($query);
            my $result = &::FetchOneColumn();
            if (!$result) {
                $query = "INSERT INTO user_group_map (user_id,group_id,isbless,isderived) " . 
                         "VALUES ($userid,$groupid,0,0)";
                &::SendSQL($query);
            }
            push (@added, $group);
        }
        if ($cmd =~ 'rem') {
            my $query = "SELECT group_id FROM user_group_map WHERE user_id = $userid " .
                        "AND group_id = $groupid AND isbless = 0 AND isderived = 0";
            &::SendSQL($query);
            my $result = &::FetchOneColumn();
            if ($result) {
                $query = "DELETE FROM user_group_map WHERE user_id = $userid " . 
                         "AND group_id = $groupid AND isbless = 0 AND isderived = 0";
                &::SendSQL($query);
            }
            push (@removed, $group);
        }
    }

    $return = "Added: " . join (" ", @added) . " " if @added;
    $return .= "Removed: " . join (" ", @removed) . " " if @removed;
    return "No actions taken. action: $cmd groups: $groups" if !$return;
    return $return;
}

### Section: Bug Query

my $querydefaults = {
    'product'                       => '',
    'version'                       => '',
    'component'                     => '',
    'bug_status'                    => [
                                        'NEW',
                                        'ASSIGNED',
                                        'REOPENED'
                                       ],
    'bug_status_list'               => [],
    'resolution'                    => '',
    'resolution_list'               => [],
    'bug_severity'                  => '',
    'bug_severity_list'             => [],
    'priority'                      => '',
    'priority_list'                 => [],
    'rep_platform'                  => '',
    'rep_platform_list'             => [],
    'op_sys'                        => '',
    'op_sys_list'                   => [],

    'short_desc_type_list'          => [
                                        'allwordssubstr',
                                        'anywordssubstr',
                                        'substring',
                                        'casesubstring',
                                        'allwords',
                                        'anywords',
                                        'regexp',
                                        'notregexp'
                                       ],
    'short_desc_type'               => 'allwordssubstr',
    'short_desc'                    => '',  

    'long_desc_type_list'           => [
                                        'allwordssubstr',
                                        'anywordssubstr',
                                        'substring',
                                        'casesubstring',
                                        'allwords',
                                        'anywords',
                                        'regexp',
                                        'notregexp'
                                       ],
    'long_desc_type'                => 'allwordssubstr',
    'long_desc'                     => '',

    'bug_file_loc_type_list'        => [
                                        'allwordssubstr',
                                        'anywordssubstr',
                                        'substring',
                                        'casesubstring',
                                        'allwords',
                                        'anywords',
                                        'regexp',
                                        'notregexp'
                                       ],
    'bug_file_loc_type'             => 'allwordssubstr',
    'bug_file_loc'                  => '',

    'status_whiteboard_type_list'   => [
                                        'allwordssubstr',
                                        'anywordssubstr',
                                        'substring',
                                        'casesubstring',
                                        'allwords',
                                        'anywords',
                                        'regexp',
                                        'notregexp'
                                       ],
    'status_whiteboard_type'        => 'allwordssubstr',
    'status_whiteboard'             => '',

    'qa_whiteboard_type_list'       => [
                                        'allwordssubstr',
                                        'anywordssubstr',
                                        'substring',
                                        'casesubstring',
                                        'allwords',
                                        'anywords',
                                        'regexp',
                                        'notregexp'
                                       ],
    'qa_whiteboard_type'            => 'allwordssubstr',
    'qa_whiteboard'                 => '',

    'devel_whiteboard_type_list'    => [
                                        'allwordssubstr',
                                        'anywordssubstr',
                                        'substring',
                                        'casesubstring',
                                        'allwords',
                                        'anywords',
                                        'regexp',
                                        'notregexp'
                                       ],
    'devel_whiteboard_type'         => 'allwordssubstr',
    'devel_whiteboard'              => '',

    'keywords_type_list'            => [
                                        'allwords',
                                        'anywords',
                                        'nowords'
                                       ],
    'keywords_type'                 => 'allwords',
    'keywords'                      => '',

    'emailreporter1'                => 0,
    'emailassigned_to1'             => 1,
    'emailcc1'                      => 0,
    'emaillongdesc1'                => 0,
    'emailqa_contact1'              => 0,
    'emailtype1_list'               => [
                                        'substring',
                                        'exact',
                                        'regexp',
                                        'notregexp'
                                       ],
    'emailtype1'                    => 'exact',
    'email1'                        => '',
    
    'emailreporter2'                => 1,
    'emailassigned_to2'             => 0,
    'emailcc2'                      => 0,
    'emaillongdesc2'                => 0,
    'emailqa_contact2'              => 0,
    'emailtype2_list'               => [
                                        'substring',
                                        'exact',
                                        'regexp',
                                        'notregexp'
                                       ],
    'emailtype2'                    => 'exact',
    'email2'                        => '',

    'bugidtype_list'                => [
                                        'include',
                                        'exclude'
                                       ],
    'bugidtype'                     => 'include',
    'bug_id'                        => '',
    
    'changedin'                     => '',
    'chfield'                       => '',
    'chfield_list'                  => [],
    'chfieldfrom'                   => '',
    'chfieldto'                     => '',
    'chfieldvalue'                  => '',
        
    'cmdtype-doit'                  => 'doit',
    'remaction'                     => 'run',
    'order_list'                    => [
                                        'Bug Number Ascending', 
                                        'Bug Number Descending', 
                                        'Importance',
                                        'Assignee', 
                                        'Last Changed'
                                       ],
    'order'                         => 'Bug Number Ascending',
    
    'column_list'                   => [
                                        'id',
                                        'alias',
                                        'opendate',
                                        'changedate',
                                        'severity',
                                        'priority',
                                        'platform',
                                        'owner',
                                        'owner_realname',
                                        'reporter',
                                        'reporter_realname',
                                        'qa_contact',
                                        'qa_contact_realname',
                                        'status',
                                        'resolution',
                                        'summary',
                                        'summaryfull',
                                        'status_whiteboard',
                                        'component',
                                        'product',
                                        'version',
                                        'os',
                                        'target_milestone',
                                        'votes',
                                        'keywords',
                                        'estimated_time',
                                        'remaining_time',
                                        'actual_time',
                                        'percentage_complete'
                                       ],
    'default_column_list'           => [],
};
 
sub getQueryInfo {
    my ($username, $password) = @_;
    my $querydata = {};
    &::ConnectToDatabase();
    Authenticate($username, $password);
    &::GetVersionTable();

    # if using groups for entry, then we don't want people to see products they
    # don't have access to. Remove them from the list.
    my @products = ();
    foreach my $p (&::GetEnterableProducts()) {
        # We build up boolean hashes in the "-set" hashes for each of these things
        # before making a list because there may be duplicates names across products.
        push @products, $p;
    }

    # @products is now all the products we are ever concerned with, as a list
    # %x_set is now a unique "list" of the relevant components/versions/tms
    @products = sort { lc($a) cmp lc($b) } @products;
                                                                                                                                           
    # Create data structures representing each product.
    for (my $i = 0; $i < @products; ++$i) {
        my $p = $products[$i];
                                                                                                                                           
        if (!$::components{$p}) {
            $::components{$p} = ['undefined'];
        }
        if (!$::versions{$p}) {
            $::versions{$p} = ['undefined'];
        }
                                                                                                                                           
        # Create hash to hold attributes for each product.
        my %product = (
            'name'       => $p,
            'components' => [ sort { lc($a) cmp lc($b) } @{$::components{$p}} ],
            'versions'   => [ sort { lc($a) cmp lc($b) } @{$::versions{$p}}   ],
        );
                                                                                                                                           
        if (Param('usetargetmilestone')) {
            $product{'milestones'} =
                          [ sort { lc($a) cmp lc($b) } @{$::target_milestone{$p}} ];
        }
                                                                                                                                           
        # Assign hash back to product array.
        $products[$i] = \%product;
    }
    $querydata->{'product'} = \@products;
    
    $querydefaults->{'bug_status_list'} = \@::legal_bug_status;
    $querydefaults->{'resolution_list'} = \@::legal_resolution;
    $querydefaults->{'bug_severity_list'} = \@::legal_severity;
    $querydefaults->{'priority_list'} = \@::legal_priority;
    $querydefaults->{'rep_platform_list'} = \@::legal_platform;
    $querydefaults->{'op_sys_list'} = \@::legal_opsys;
    $querydefaults->{'default_column_list'} = \@::default_column_list;                                                                                                                                   
    my @chfields;
                                                                                                                                           
    push @chfields, "[Bug creation]";
                                                                                                                                           
    # This is what happens when you have variables whose definition depends
    # on the DB schema, and then the underlying schema changes...
    foreach my $val (@::log_columns) {
        if ($val eq 'product_id') {
            $val = 'product';
        } elsif ($val eq 'component_id') {
            $val = 'component';
        }
        push @chfields, $val;
    }

    if (&::UserInGroup(Param('timetrackinggroup'))) {
        push @chfields, "work_time";
    } else {
        @chfields = grep($_ ne "estimated_time", @chfields);
        @chfields = grep($_ ne "remaining_time", @chfields);
    }
    @chfields = (sort(@chfields));
    $querydefaults->{'chfield_list'} = \@chfields;
    
    $querydata->{'userid'} = $::userid;
    $querydata->{'columns'} = \@::log_columns;                                                                                                                                       
    # Boolean charts
    my @fields;
    push(@fields, { name => "noop", description => "---" });
    &::SendSQL("SELECT name, description FROM fielddefs ORDER BY sortkey");
    while (&::MoreSQLData()) {
        my ($name, $description) = &::FetchSQLData();
        if (($name eq "estimated_time" ||
            $name eq "remaining_time" ||
            $name eq "work_time" ||
            $name eq "percentage_complete" ) &&
            (!&::UserInGroup(Param('timetrackinggroup')))) {
            next;
        }
        push(@fields, { name => $name, description => $description });
    }
    $querydata->{'fields'} = \@fields;

    return [$querydata, $querydefaults];
}

my $columns = {};
sub DefineColumn {
    my ($id, $name, $title) = @_;
    $columns->{$id} = { 'name' => $name , 'title' => $title };
}

sub runQuery {
    my ($querydata, $username, $password) = @_;
    &::ConnectToDatabase();
    Authenticate($username, $password);
  
    if (ref($querydata) !~ /HASH/) {
        $RPC_FAULT_CODE = 5;
        $RPC_FAULT_MESSAGE = "No query information given";
        return;
    }

	# Column:     ID                    Name                           Title
	DefineColumn("id"                , "bugs.bug_id"                , "ID"               );
    DefineColumn("alias"             , "bugs.alias"                 , "Alias"            );
	DefineColumn("opendate"          , "bugs.creation_ts"           , "Opened"           );
	DefineColumn("changeddate"       , "bugs.delta_ts"              , "Changed"          );
	DefineColumn("severity"          , "bugs.bug_severity"          , "Severity"         );
	DefineColumn("priority"          , "bugs.priority"              , "Priority"         );
	DefineColumn("platform"          , "bugs.rep_platform"          , "Platform"         );
	DefineColumn("owner"             , "map_assigned_to.login_name" , "Owner"            );
	DefineColumn("owner_realname"    , "map_assigned_to.realname"   , "Owner"            );
	DefineColumn("reporter"          , "map_reporter.login_name"    , "Reporter"         );
	DefineColumn("reporter_realname" , "map_reporter.realname"      , "Reporter"         );
	DefineColumn("qa_contact"        , "map_qa_contact.login_name"  , "QA Contact"       );
	DefineColumn("qa_contact_realname", "map_qa_contact.realname"   , "QA Contact"       );
	DefineColumn("status"            , "bugs.bug_status"            , "State"            );
	DefineColumn("resolution"        , "bugs.resolution"            , "Result"           );
	DefineColumn("summary"           , "bugs.short_desc"            , "Summary"          );
	DefineColumn("summaryfull"       , "bugs.short_desc"            , "Summary"          );
	DefineColumn("status_whiteboard" , "bugs.status_whiteboard"     , "Status Summary"   );
    DefineColumn("devel_whiteboard"  , "bugs.devel_whiteboard"      , "Devel  Summary"   );
    DefineColumn("qa_whiteboard"     , "bugs.qa_whiteboard"         , "QA Status Summary");
	DefineColumn("component"         , "map_components.name"        , "Component"        );
	DefineColumn("product"           , "map_products.name"          , "Product"          );
	DefineColumn("version"           , "bugs.version"               , "Version"          );
	DefineColumn("os"                , "bugs.op_sys"                , "OS"               );
	DefineColumn("target_milestone"  , "bugs.target_milestone"      , "Target Milestone" );
	DefineColumn("votes"             , "bugs.votes"                 , "Votes"            );
    DefineColumn("keywords"          , "bugs.keywords"              , "Keywords"         );
    DefineColumn("estimated_time"   , "bugs.estimated_time"       , "Estimated Hours"  );
    DefineColumn("remaining_time"   , "bugs.remaining_time"       , "Remaining Hours"  );
    DefineColumn("actual_time"      , "(SUM(ldtime.work_time)*COUNT(DISTINCT ldtime.bug_when)/COUNT(bugs.bug_id)) AS actual_time", "Actual Hours");
    DefineColumn("percentage_complete","(100*((SUM(ldtime.work_time)*COUNT(DISTINCT ldtime.bug_when)/COUNT(bugs.bug_id))/((SUM(ldtime.work_time)*COUNT(DISTINCT ldtime.bug_when)/COUNT(bugs.bug_id))+bugs.remaining_time))) AS percentage_complete", "% Complete");
 
    my @displaycolumns;
    if (defined $querydata->{'column_list'} && @{$querydata->{'column_list'}} > 0) {
        @displaycolumns = @{$querydata->{'column_list'}};
    } else {
        @displaycolumns =  @::default_column_list;
    }

    # Remove the timetracking columns if they are not a part of the group
    # (happens if a user had access to time tracking and it was revoked/disabled)
    if (!&::UserInGroup(Param("timetrackinggroup"))) {
        @displaycolumns = grep($_ ne 'estimated_time', @displaycolumns);
        @displaycolumns = grep($_ ne 'remaining_time', @displaycolumns);
        @displaycolumns = grep($_ ne 'actual_time', @displaycolumns);
        @displaycolumns = grep($_ ne 'percentage_complete', @displaycolumns);
    }

    my @selectcolumns = ("id");

    # remaining and actual_time are required for precentage_complete calculation:
    if (lsearch(\@displaycolumns, "percentage_complete") >= 0) {
        push (@selectcolumns, "remaining_time");
        push (@selectcolumns, "actual_time");
    }

    # Display columns are selected because otherwise we could not display them.
    push (@selectcolumns, @displaycolumns);
    
    # Convert the list of columns being selected into a list of column names.
    my @selectnames = map($columns->{$_}->{'name'}, @selectcolumns);

    # Remove columns with no names, such as percentage_complete
    #  (or a removed *_time column due to permissions)
    @selectnames = grep($_ ne '', @selectnames);
  
    my $order = $querydata->{'order'} || "Bug Number Ascending";
 
    my $db_order = "";  # Modified version of $order for use with SQL query 
    # Convert the value of the "order" form field into a list of columns
    # by which to sort the results.
    ORDER: for ($order) {
        /\./ && do {
            my @columnnames = map($columns->{lc($_)}->{'name'}, keys(%$columns));
            # A custom list of columns.  Make sure each column is valid.
            foreach my $fragment (split(/,/, $order)) {
                $fragment = trim($fragment);
                # Accept an order fragment matching a column name, with
                # asc|desc optionally following (to specify the direction)
                if (!grep($fragment =~ /^\Q$_\E(\s+(asc|desc))?$/, @columnnames)) {
                    $vars->{'fragment'} = $fragment;
                    ThrowCodeError("invalid_column_name_form");
                }
            }
            # Now that we have checked that all columns in the order are valid,
            # detaint the order string.
            trick_taint($order);
            last ORDER; 
        };
        /Ascending/ && do {
            $order = "bugs.bug_id asc";
            last ORDER;
        };
        /Descending/ && do {
            $order = "bugs.bug_id desc";
            last ORDER;
        };
        /Import/ && do {
            $order = "bugs.priority, bugs.bug_severity";
            last ORDER;
        };
        /Assign/ && do {
            $order = "map_assigned_to.login_name, bugs.bug_status, bugs.priority, bugs.bug_id";
            last ORDER;
        };
        /Changed/ && do {
            $order = "bugs.delta_ts, bugs.bug_status, bugs.priority, map_assigned_to.login_name, bugs.bug_id";
            last ORDER;
        };
        # DEFAULT
        $order = "bugs.bug_status, bugs.priority, map_assigned_to.login_name, bugs.bug_id";
    }

    foreach my $fragment (split(/,/, $order)) {
        $fragment = trim($fragment);
        if (!grep($fragment =~ /^\Q$_\E(\s+(asc|desc))?$/, @selectnames)) {
            # Add order columns to selectnames
            # The fragment has already been validated
            $fragment =~ s/\s+(asc|desc)$//;
            $fragment =~ tr/a-zA-Z\.0-9\-_//cd;
            push @selectnames, $fragment;
        }
    }
    $db_order = $order;  # Copy $order into $db_order for use with SQL query
                                                                                                                                           
    # If we are sorting by votes, sort in descending order if no explicit
    # sort order was given
    $db_order =~ s/bugs.votes\s*(,|$)/bugs.votes desc$1/i;
                                                                                                                                           
    # the 'actual_time' field is defined as an aggregate function, but
    # for order we just need the column name 'actual_time'
    my $aggregate_search = quotemeta($columns->{'actual_time'}->{'name'});
    $db_order =~ s/$aggregate_search/actual_time/g;
                                                                                                                                           
    # the 'percentage_complete' field is defined as an aggregate too
    $aggregate_search = quotemeta($columns->{'percentage_complete'}->{'name'});
    $db_order =~ s/$aggregate_search/percentage_complete/g;
  
    use Bugzilla::CGI;
    use Bugzilla::Search;
 
    my $params = new Bugzilla::CGI;

    foreach my $default_key (keys %{$querydefaults}) {
        next if $default_key =~ /_list/;
        if (!$querydata->{$default_key}) {
            if ((ref($querydefaults->{$default_key}) =~ /ARRAY/ && @{$querydefaults->{$default_key}} > 0) || $querydefaults->{$default_key} ne '') {
                $params->param(-name=>$default_key, -value=>$querydefaults->{$default_key}); 
            } 
        } else {
            $params->param(-name=>$default_key, -value=>$querydata->{$default_key});
        }
    }

    my $search = new Bugzilla::Search('fields' => \@selectnames,
                                      'params' => $params);
    my $query = $search->getSQL();

    # Extra special disgusting hack: if we are ordering by target_milestone,
    # change it to order by the sortkey of the target_milestone first.
    if ($db_order =~ /bugs.target_milestone/) {
        $db_order =~ s/bugs.target_milestone/ms_order.sortkey,ms_order.value/;
        $query =~ s/\sWHERE\s/ LEFT JOIN milestones ms_order ON ms_order.value = bugs.target_milestone AND ms_order.product_id = bugs.product_id WHERE /;
    }
                                                                                                                                           
    $query .= " ORDER BY $db_order " if ($order);

    # Normally, we ignore SIGTERM and SIGPIPE (see globals.pl) but we need to
    # respond to them here to prevent someone DOSing us by reloading a query
    # a large number of times.
    $::SIG{TERM} = 'DEFAULT';
    $::SIG{PIPE} = 'DEFAULT';

    # Execute the query.
    &::SendSQL($query);

    # Retrieve the query results one row at a time and write the data into a list
    # of Perl records.
                                                                                                                                           
    my $bugowners = {};
    my $bugproducts = {};
    my $bugstatuses = {};
    my @bugidlist;
                                                                                                                                           
    my @bugs; # the list of records
                                                                                                                                           
    while (my @row = &::FetchSQLData()) {
        my $bug = {}; # a record
                                                                                                                                           
        # Slurp the row of data into the record.
        # The second from last column in the record is the number of groups
        # to which the bug is restricted.
        foreach my $column (@selectcolumns) {
            $bug->{$column} = shift @row;
        }
                                                                                                                                           
        # Process certain values further (i.e. date format conversion).
        if ($bug->{'changeddate'}) {
            if ($::db_driver eq 'mysql') {
                $bug->{'changeddate'} =~
                s/^(\d{4})(\d{2})(\d{2})(\d{2})(\d{2})(\d{2})$/$1-$2-$3 $4:$5:$6/;
            } elsif ($::db_driver eq 'Pg') {
                $bug->{'changeddate'} =~
                s/^(\d{4})-(\d{2})-(\d{2}) (\d{2}):(\d{2}):(\d{2,2}).*/$1-$2-$3 $4:$5:$6/;
            }
            $bug->{'opendate'} = DiffDate($bug->{'opendate'});
        }
                                                                                                                                           
        # Record the owner, product, and status in the big hashes of those things.
        $bugowners->{$bug->{'owner'}} = 1 if $bug->{'owner'};
        $bugproducts->{$bug->{'product'}} = 1 if $bug->{'product'};
        $bugstatuses->{$bug->{'status'}} = 1 if $bug->{'status'};

        $bug->{isingroups} = 0;

        # Add the record to the list.
        push(@bugs, $bug);

        # Add id to list for checking for bug privacy later
        push(@bugidlist, $bug->{id});
    }

    # Switch back from the shadow database to the regular database so PutFooter()
    # can determine the current user even if the "logincookies" table is corrupted
    # in the shadow database.
    &::ReconnectToMainDatabase();
                                                                                                                                           
    # Check for bug privacy and set $bug->{isingroups} = 1 if private
    # to 1 or more groups
    my %privatebugs;
    if (@bugidlist && @bugidlist < 2000) {
        &::SendSQL("SELECT DISTINCT bugs.bug_id FROM bugs, bug_group_map " .
                "WHERE bugs.bug_id = bug_group_map.bug_id " .
                "AND bugs.bug_id IN (" . join(',',@bugidlist) . ")");
        while (&::MoreSQLData()) {
            my ($id) = &::FetchSQLData();
            $privatebugs{$id} = 1;
        }
        foreach my $bug (@bugs) {
            if ($privatebugs{$bug->{id}}) {
                $bug->{isingroups} = 1;
            }
        }
    }
    
    my $bugdata = {};
    $bugdata->{'bugs'} = \@bugs;
    $bugdata->{'displaycolumns'} = \@displaycolumns;
    # $bugdata->{'sql'} = $query;

    return $bugdata;
}

sub DiffDate {
    my ($datestr) = @_;
    my @weekday= qw( Sun Mon Tue Wed Thu Fri Sat );

    my $date = str2time($datestr);
    my $age = time() - $date;
    my ($s,$m,$h,$d,$mo,$y,$wd)= localtime $date;
    if( $age < 18*60*60 ) {
        $date = sprintf "%02d:%02d:%02d", $h,$m,$s;
    } elsif( $age < 6*24*60*60 ) {
        $date = sprintf "%s %02d:%02d", $weekday[$wd],$h,$m;
    } else {
        $date = sprintf "%04d-%02d-%02d", 1900+$y,$mo+1,$d;
    }
    return $date;
}


1;
